# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
A general tokenizer
"""

import collections
import re

Token = collections.namedtuple("Token", ["kind", "value"])


def new_token_kind(name):
    """
    Create a new token kind with nice __repr__
    """
    cls = type(name, (object,), {"__repr__": lambda self: name})
    return cls()


class Tokenizer(object):
    """
    Maintain a prioritized list of token regex
    """

    def __init__(self):
        self._regexs = []
        self._assoc = {}
        self._regex = None

    def add(self, name, regex, func=None):
        """
        Add token type
        """
        key = chr(ord('a') + len(self._regexs))
        self._regexs.append((key, regex))
        kind = new_token_kind(name)
        self._assoc[key] = (kind, func)
        return kind

    def finalize(self):
        self._regex = re.compile("|".join("(?P<%s>%s)" % spec for spec in self._regexs), re.VERBOSE | re.MULTILINE)

    def tokenize(self, code):
        """
        Tokenize the code
        """
        tokens = []
        start = 0
        while True:
            match = self._regex.search(code, pos=start)
            if match is None:
                break
            start = match.end()
            key = match.lastgroup
            kind, func = self._assoc[key]
            value = match.group(match.lastgroup)
            token = Token(kind, value)
            if func is not None:
                token = func(token)

            if token is not None:
                tokens.append(token)
        return tokens


class TokenStream(object):
    """
    Helper class for traversing a stream of tokens
    """

    def __init__(self, tokens):
        self._tokens = tokens
        self._idx = 0

    def __len__(self):
        return len(self._tokens)

    def __getitem__(self, index):
        return self._tokens[index]

    @property
    def eof(self):
        return not self._idx < len(self._tokens)

    @property
    def idx(self):
        return self._idx

    @property
    def current(self):
        return self._tokens[self._idx]

    def peek(self, offset=0):
        return self._tokens[self._idx + offset]

    def skip_while(self, kind):
        """
        Skip forward while token kind is present
        """
        while (not self.eof) and self._tokens[self._idx].kind == kind:
            self._idx += 1
        return self._idx

    def skip_until(self, kind):
        """
        Skip forward until token kind is present
        """
        while (not self.eof) and self._tokens[self._idx].kind != kind:
            self._idx += 1
        return self._idx

    def pop(self):
        """
        Return current token and advance stream
        """
        self._idx += 1
        return self._tokens[self._idx - 1]

    def slice(self, start, end):
        return self._tokens[start:end]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Verilog parsing functionality
"""
from os.path import join, exists

from vunit.parsing.tokenizer import TokenStream
import vunit.parsing.verilog.tokenizer as tokenizer
from vunit.parsing.verilog.tokenizer import tokenize


def preprocess(tokens, defines=None, include_paths=None, included_files=None):
    """
    Pre-process tokens while filling in defines
    """
    stream = TokenStream(tokens)
    include_paths = [] if include_paths is None else include_paths
    included_files = [] if included_files is None else included_files
    defines = {} if defines is None else defines
    result = []

    while not stream.eof:
        token = stream.pop()
        if not token.kind == tokenizer.PREPROCESSOR:
            result.append(token)
            continue

        if token.value == "define":
            macro = define(stream)
            defines[macro.name] = macro

        if token.value == "include":
            stream.skip_until(tokenizer.STRING)
            file_name = stream.pop().value

            full_name = None
            for include_path in include_paths:
                full_name = join(include_path, file_name)
                if exists(full_name):
                    break
            else:
                assert False
            included_files.append(full_name)
            with open(full_name, "r") as fptr:
                included_tokens = tokenize(fptr.read())
            result += preprocess(included_tokens, defines, include_paths, included_files)

        elif token.value in defines:
            macro = defines[token.value]
            if macro.num_args == 0:
                values = []
            else:
                values = parse_macro_actuals(stream)
            result += macro.expand(values)

    return result


def define(stream):
    """
    Handle a `define directive
    """
    stream.skip_while(tokenizer.WHITESPACE)
    name_token = stream.pop()
    assert name_token.kind == tokenizer.IDENTIFIER
    name = name_token.value

    if stream.eof:
        return Macro(name)

    token = stream.pop()
    if token.kind in (tokenizer.WHITESPACE, tokenizer.NEWLINE):
        # Define without arguments
        args = tuple()
        defaults = {}
    elif token.kind == tokenizer.LPAR:
        args = tuple()
        defaults = {}
        while token.kind != tokenizer.RPAR:
            if token.kind == tokenizer.IDENTIFIER:
                argname = token.value
                args = args + (argname,)
                token = stream.pop()
                if token.kind == tokenizer.EQUAL:
                    token = stream.pop()
                    defaults[argname] = [token]
                    token = stream.pop()
            else:
                token = stream.pop()

    start = stream.idx
    end = stream.skip_until(tokenizer.NEWLINE)
    if not stream.eof:
        stream.pop()
    return Macro(name,
                 tokens=stream.slice(start, end),
                 args=args,
                 defaults=defaults)


def parse_macro_actuals(stream):
    """
    Parse the actual values of macro call such as
    1 2 in `macro(1, 2)
    """
    token = stream.pop()
    assert token.kind == tokenizer.LPAR
    token = stream.pop()

    value = []
    values = []
    while token.kind != tokenizer.RPAR:
        if token.kind == tokenizer.COMMA:
            values.append(value)
            value = []
        else:
            value.append(token)
        token = stream.pop()
    values.append(value)
    return values


class Macro(object):
    """
    A `define macro with zero or more arguments
    """

    def __init__(self, name, tokens=None, args=tuple(), defaults=None):
        self.name = name
        self.tokens = [] if tokens is None else tokens
        self.args = args
        self.defaults = {} if defaults is None else defaults

    @property
    def num_args(self):
        return len(self.args)

    def __repr__(self):
        return "Macro(%r, %r %r, %r)" % (self.name, self.tokens, self.args, self.defaults)

    def expand(self, values):
        """
        Expand macro with actual values, returns a list of expanded tokens
        """
        tokens = []
        for token in self.tokens:
            if token.kind == tokenizer.IDENTIFIER and token.value in self.args:
                idx = self.args.index(token.value)
                if idx >= len(values):
                    value = self.defaults[token.value]
                else:
                    value = values[idx]
                tokens += value
            else:
                tokens.append(token)
        return tokens

    def __eq__(self, other):
        return ((self.name == other.name) and
                (self.tokens == other.tokens) and
                (self.args == other.args) and
                (self.defaults == other.defaults))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Verilog parsing functionality
"""

from os.path import dirname
from vunit.parsing.tokenizer import TokenStream
from vunit.parsing.verilog.tokenizer import tokenize
from vunit.parsing.verilog.preprocess import preprocess
import vunit.parsing.verilog.tokenizer as tokenizer


class VerilogParser(object):
    """
    Parse a single Verilog file
    """

    def __init__(self):
        pass

    @staticmethod
    def parse(code, file_name, include_paths, content_hash):  # pylint: disable=unused-argument
        return VerilogDesignFile.parse(code, file_name, include_paths)


class VerilogDesignFile(object):
    """
    Contains Verilog objecs found within a file
    """
    def __init__(self,  # pylint: disable=too-many-arguments
                 modules=None,
                 packages=None,
                 imports=None,
                 instances=None,
                 included_files=None):
        self.modules = [] if modules is None else modules
        self.packages = [] if packages is None else packages
        self.imports = [] if imports is None else imports
        self.instances = [] if instances is None else instances
        self.included_files = [] if included_files is None else included_files

    @classmethod
    def parse(cls, code, file_name, include_paths=None):
        """
        Parse verilog file
        """
        include_paths = [] if include_paths is None else include_paths
        tokens = tokenize(code)
        included_files = []
        pp_tokens = preprocess(tokens, include_paths=[dirname(file_name)] + include_paths,
                               included_files=included_files)
        tokens = [token
                  for token in pp_tokens
                  if token.kind not in (tokenizer.WHITESPACE,
                                        tokenizer.COMMENT,
                                        tokenizer.NEWLINE,
                                        tokenizer.MULTI_COMMENT)]
        return cls(modules=VerilogModule.find(tokens),
                   packages=VerilogPackage.find(tokens),
                   imports=cls.find_imports(tokens),
                   instances=cls.find_instances(tokens),
                   included_files=included_files)

    @staticmethod
    def find_imports(tokens):
        """
        Find imports
        """
        results = []
        stream = TokenStream(tokens)
        while not stream.eof:
            token = stream.pop()
            if token.kind == tokenizer.IMPORT:
                stream.skip_until(tokenizer.IDENTIFIER)
                name = stream.pop().value
                results.append(name)
        return results

    @staticmethod
    def find_instances(tokens):
        """
        Find module instances
        """
        results = []
        stream = TokenStream(tokens)
        while not stream.eof:
            token = stream.pop()

            if not token.kind == tokenizer.IDENTIFIER:
                continue
            modulename = token.value

            token = stream.pop()
            if token.kind == tokenizer.HASH:
                results.append(modulename)
                continue
            elif token.kind == tokenizer.IDENTIFIER:
                results.append(modulename)
                continue

        return results


class VerilogModule(object):
    """
    A verilog module
    """

    def __init__(self, name, parameters):
        self.name = name
        self.parameters = parameters

    @classmethod
    def parse_parameter(cls, idx, tokens):
        """
        Parse parameter at point
        """
        if not tokens[idx].kind == tokenizer.PARAMETER:
            return None

        if tokens[idx + 2].kind == tokenizer.IDENTIFIER:
            return tokens[idx + 2].value
        else:
            return tokens[idx + 1].value
        assert False

    @classmethod
    def find(cls, tokens):
        """
        Find all modules within code, nested modules are ignored
        """
        idx = 0
        name = None
        balance = 0
        results = []
        parameters = []
        while idx < len(tokens):

            if tokens[idx].kind == tokenizer.MODULE:
                if balance == 0:
                    name = tokens[idx + 1].value
                    parameters = []
                balance += 1

            elif tokens[idx].kind == tokenizer.ENDMODULE:
                balance -= 1
                if balance == 0:
                    results.append(cls(name, parameters))

            elif balance == 1:
                parameter = cls.parse_parameter(idx, tokens)
                if parameter is not None:
                    parameters.append(parameter)

            idx += 1
        return results


class VerilogPackage(object):
    """
    A verilog package
    """

    def __init__(self, name):
        self.name = name

    @classmethod
    def find(cls, tokens):
        """
        Find all modules within code, nested modules are ignored
        """
        idx = 0
        results = []
        while idx < len(tokens):
            if tokens[idx].kind == tokenizer.PACKAGE:
                idx += 1
                name = tokens[idx].value
                results.append(cls(name))
            idx += 1
        return results



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Test of the Verilog parser
"""

from unittest import TestCase
from vunit.parsing.verilog.parser import VerilogDesignFile


class TestVerilogParser(TestCase):  # pylint: disable=too-many-public-methods
    """
    Test of the Verilog parser
    """

    def test_parsing_empty(self):
        design_file = parse("")
        self.assertEqual(design_file.modules, [])

    def test_parse_module(self):
        modules = parse("""\
module true1;
  my_module hello
  "module false";
endmodule
/*
module false
*/
  module true2; // module false
endmodule module true3
endmodule
""").modules
        self.assertEqual(len(modules), 3)
        self.assertEqual(modules[0].name, "true1")
        self.assertEqual(modules[1].name, "true2")
        self.assertEqual(modules[2].name, "true3")

    def test_parse_parameter_without_type(self):
        modules = parse("""\
module foo;
  parameter param1;
  parameter param2 = 1;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_parse_parameter_with_type(self):
        modules = parse("""\
module foo;
  parameter string param1;
  parameter integer param2 = 1;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_nested_modules_are_ignored(self):
        modules = parse("""\
module foo;
  parameter string param1;
  module nested;
    parameter integer param_nested;
  endmodule
  parameter string param2;
endmodule
""").modules
        self.assertEqual(len(modules), 1)
        module = modules[0]
        self.assertEqual(module.name, "foo")
        self.assertEqual(len(module.parameters), 2)
        param1, param2 = module.parameters
        self.assertEqual(param1, "param1")
        self.assertEqual(param2, "param2")

    def test_parse_package(self):
        packages = parse("""\
package true1;
endpackage package true2; endpackage
""").packages
        self.assertEqual(len(packages), 2)
        self.assertEqual(packages[0].name, "true1")
        self.assertEqual(packages[1].name, "true2")

    def test_parse_imports(self):
        imports = parse("""\
import true1;
package pkg;
  import true2::*;
endpackage
""").imports
        self.assertEqual(len(imports), 2)
        self.assertEqual(imports[0], "true1")
        self.assertEqual(imports[1], "true2")

    def test_parse_instances(self):
        instances = parse("""\
module name;
  true1 instance_name1();
  true2 instance_name2(.foo(bar));
  true3 #(.param(1)) instance_name3(.foo(bar));
endmodule
""").instances
        self.assertEqual(len(instances), 3)
        self.assertEqual(instances[0], "true1")
        self.assertEqual(instances[1], "true2")
        self.assertEqual(instances[2], "true3")


def parse(code):
    """
    Helper function to parse
    """
    design_file = VerilogDesignFile.parse(code, "file_name.sv", [])
    return design_file



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
The main public Python interface of VUnit.
"""


from __future__ import print_function

import sys
import os
import traceback

from os.path import exists, relpath, abspath, join, basename, splitext
from glob import glob
from fnmatch import fnmatch
from vunit.database import PickledDataBase, DataBase

import vunit.ostools as ostools
from vunit.vunit_cli import VUnitCLI
from vunit.simulator_factory import SimulatorFactory
from vunit.color_printer import (COLOR_PRINTER,
                                 NO_COLOR_PRINTER)
from vunit.project import Project, file_type_of
from vunit.test_runner import TestRunner
from vunit.test_report import TestReport
from vunit.test_scanner import TestScanner, TestScannerError, tb_filter
from vunit.test_configuration import TestConfiguration, create_scope
from vunit.exceptions import CompileError
from vunit.location_preprocessor import LocationPreprocessor
from vunit.check_preprocessor import CheckPreprocessor
from vunit.vhdl_parser import CachedVHDLParser
from vunit.builtins import (add_vhdl_builtins,
                            add_verilog_include_dir,
                            add_array_util,
                            add_osvvm,
                            add_com)
from vunit.com import codec_generator

import logging
LOGGER = logging.getLogger(__name__)


class VUnit(object):  # pylint: disable=too-many-instance-attributes, too-many-public-methods
    """
    The public interface of VUnit
    """

    @classmethod
    def from_argv(cls, argv=None, compile_builtins=True):
        """
        Create VUnit instance from command line arguments
        Can take arguments from 'argv' if not None  instead of sys.argv
        """
        args = VUnitCLI().parse_args(argv=argv)
        return cls.from_args(args, compile_builtins=compile_builtins)

    @classmethod
    def from_args(cls, args, compile_builtins=True):
        """
        Create VUnit instance from args namespace
        """
        def test_filter(name):
            return any(fnmatch(name, pattern) for pattern in args.test_patterns)

        return cls(output_path=args.output_path,
                   clean=args.clean,
                   vhdl_standard=select_vhdl_standard(),
                   use_debug_codecs=args.use_debug_codecs,
                   no_color=args.no_color,
                   verbose=args.verbose,
                   xunit_xml=args.xunit_xml,
                   log_level=args.log_level,
                   test_filter=test_filter,
                   list_only=args.list,
                   compile_only=args.compile,
                   elaborate_only=args.elaborate,
                   compile_builtins=compile_builtins,
                   simulator_factory=SimulatorFactory(args),
                   num_threads=args.num_threads,
                   exit_0=args.exit_0)

    def __init__(self,  # pylint: disable=too-many-locals, too-many-arguments
                 output_path,
                 simulator_factory,
                 clean=False,
                 use_debug_codecs=False,
                 no_color=False,
                 verbose=False,
                 xunit_xml=None,
                 log_level="warning",
                 test_filter=None,
                 list_only=False,
                 compile_only=False,
                 elaborate_only=False,
                 vhdl_standard='2008',
                 compile_builtins=True,
                 num_threads=1,
                 exit_0=False):

        self._configure_logging(log_level)

        self._output_path = output_path

        if no_color:
            self._printer = NO_COLOR_PRINTER
        else:
            self._printer = COLOR_PRINTER

        self._verbose = verbose
        self._xunit_xml = xunit_xml

        self._test_filter = test_filter if test_filter is not None else lambda name: True
        self._list_only = list_only
        self._compile_only = compile_only
        self._vhdl_standard = vhdl_standard

        self._tb_filter = tb_filter
        self._configuration = TestConfiguration(elaborate_only=elaborate_only)
        self._external_preprocessors = []
        self._location_preprocessor = None
        self._check_preprocessor = None
        self._use_debug_codecs = use_debug_codecs

        self._simulator_factory = simulator_factory
        self._create_output_path(clean)

        self._project = None
        self._create_project()
        self._num_threads = num_threads
        self._exit_0 = exit_0

        if compile_builtins:
            self.add_builtins(library_name="vunit_lib")

    def _create_project(self):
        """
        Create Project instance
        """
        database = self._create_database()
        self._project = Project(
            vhdl_parser=CachedVHDLParser(database=database),
            depend_on_package_body=self._simulator_factory.package_users_depend_on_bodies())

    def _create_database(self):
        """
        Create a persistent database to store expensive parse results

        Check for Python version used to create the database is the
        same as the running python instance or re-create
        """
        project_database_file_name = join(self._output_path, "project_database")
        create_new = False
        key = b"version"
        version = str((5, sys.version)).encode()
        database = None
        try:
            database = DataBase(project_database_file_name)
            create_new = (key not in database) or (database[key] != version)
        except KeyboardInterrupt:
            raise
        except:  # pylint: disable=bare-except
            traceback.print_exc()
            create_new = True

        if create_new:
            database = DataBase(project_database_file_name, new=True)
        database[key] = version

        return PickledDataBase(database)

    @staticmethod
    def _configure_logging(log_level):
        """
        Configure logging based on log_level string
        """
        level = getattr(logging, log_level.upper())
        logging.basicConfig(filename=None, format='%(levelname)7s - %(message)s', level=level)

    def add_external_library(self, library_name, path):
        """
        Add external black box library
        """
        self._project.add_library(library_name, abspath(path), is_external=True)
        return self._create_library_facade(library_name)

    def add_library(self, library_name):
        """
        Add vunit managed white box library
        """
        path = join(self._simulator_factory.simulator_output_path, "libraries", library_name)
        self._project.add_library(library_name, abspath(path))
        return self._create_library_facade(library_name)

    def library(self, library_name):
        """
        Get reference to library
        """
        if not self._project.has_library(library_name):
            raise KeyError(library_name)
        return self._create_library_facade(library_name)

    def _create_library_facade(self, library_name):
        """
        Create a Library object to be exposed to users
        """
        return LibraryFacade(library_name, self, self._project, self._configuration)

    def set_generic(self, name, value):
        """
        Globally set generic
        """
        self._configuration.set_generic(name.lower(), value, scope=create_scope())

    def set_parameter(self, name, value):
        """
        Globally set parameter
        """
        self.set_generic(name, value)

    def set_sim_option(self, name, value):
        """
        Globally set simulation option
        """
        self._configuration.set_sim_option(name, value, scope=create_scope())

    def set_compile_option(self, name, value):
        """
        Globally set compile option
        """
        for source_file in self._project.get_source_files_in_order():
            source_file.set_compile_option(name, value)

    def set_pli(self, value):
        """
        Globally set pli
        """
        self._configuration.set_pli(value, scope=create_scope())

    def disable_ieee_warnings(self):
        """
        Globally disable ieee warnings
        """
        self._configuration.disable_ieee_warnings(scope=create_scope())

    def get_source_file(self, file_name, library_name=None):
        """
        Get specific source file in specfic or any library.
        Error if multiple matches.
        """

        files = self.get_source_files(file_name, library_name, allow_empty=True)
        if len(files) > 1:
            raise ValueError("Found file named '%s' in multiple-libraries, "
                             "add explicit library_name." % file_name)
        elif len(files) == 0:
            if library_name is None:
                raise ValueError("Found no file named '%s'" % file_name)
            else:
                raise ValueError("Found no file named '%s' in library '%s'"
                                 % (file_name, library_name))
        return files[0]

    def get_source_files(self, pattern="*", library_name=None, allow_empty=False):
        """
        Get source files matching wildcard pattern in either a specific
        library or all libraries
        """
        results = []
        for source_file in self._project.get_source_files_in_order():
            if library_name is not None:
                if source_file.library.name != library_name:
                    continue

            if not (fnmatch(abspath(source_file.name), pattern) or
                    fnmatch(relpath(source_file.name), pattern)):
                continue

            results.append(FileFacade(source_file, self._project))

        if (not allow_empty) and len(results) == 0:
            raise ValueError(("Pattern %r did not match any file. "
                              "Use allow_empty=True to avoid exception,") % pattern)

        return FileSetFacade(results)

    def add_source_files(self,   # pylint: disable=too-many-arguments
                         pattern, library_name, preprocessors=None, include_dirs=None, allow_empty=False):
        """
        Add source files matching wildcard pattern to library
        """
        file_names = glob(pattern)

        if (not allow_empty) and len(file_names) == 0:
            raise ValueError(("Pattern %r did not match any file. "
                              "Use allow_empty=True to avoid exception,") % pattern)

        return FileSetFacade(source_files=[
            self.add_source_file(file_name, library_name, preprocessors, include_dirs)
            for file_name in file_names])

    def add_source_file(self, file_name, library_name, preprocessors=None, include_dirs=None):
        """
        Add source file to library
        """
        file_type = file_type_of(file_name)

        if file_type == "verilog":
            include_dirs = include_dirs if include_dirs is not None else []
            add_verilog_include_dir(include_dirs)

        file_name = self._preprocess(library_name, abspath(file_name), preprocessors)
        return FileFacade(self._project.add_source_file(file_name,
                                                        library_name,
                                                        file_type=file_type,
                                                        include_dirs=include_dirs),
                          self._project)

    def _preprocess(self, library_name, file_name, preprocessors):
        """
        Preprocess file_name within library_name using explicit preprocessors
        if preprocessors is None then use implicit globally defined processors
        """
        # @TODO dependency checking etc...

        if preprocessors is None:
            preprocessors = [self._location_preprocessor, self._check_preprocessor]
            preprocessors = [p for p in preprocessors if p is not None]
            preprocessors = self._external_preprocessors + preprocessors

        if len(preprocessors) == 0:
            return file_name

        code = ostools.read_file(file_name)
        for preprocessor in preprocessors:
            code = preprocessor.run(code, basename(file_name))

        pp_file_name = join(self._preprocessed_path, library_name, basename(file_name))

        idx = 1
        while ostools.file_exists(pp_file_name):
            LOGGER.debug("Preprocessed file exists '%s', adding prefix", pp_file_name)
            pp_file_name = join(self._preprocessed_path,
                                library_name, "%i_%s" % (idx, basename(file_name)))
            idx += 1

        ostools.write_file(pp_file_name, code)
        return pp_file_name

    def add_preprocessor(self, preprocessor):
        """
        Add a custom preprocessor to be used on all files, must be called before adding any files
        """
        self._external_preprocessors.append(preprocessor)

    def enable_location_preprocessing(self, additional_subprograms=None):
        """
        Enable location preprocessing, must be called before adding any files
        """
        preprocessor = LocationPreprocessor()
        if additional_subprograms is not None:
            for subprogram in additional_subprograms:
                preprocessor.add_subprogram(subprogram)
        self._location_preprocessor = preprocessor

    def enable_check_preprocessing(self):
        """
        Enable check preprocessing, must be called before adding any files
        """
        self._check_preprocessor = CheckPreprocessor()

    def main(self):
        """
        Run vunit main function and exit with code
        """
        try:
            all_ok = self._main()
        except KeyboardInterrupt:
            exit(1)
        except CompileError:
            exit(1)
        except TestScannerError:
            exit(1)
        except:  # pylint: disable=bare-except
            traceback.print_exc()
            exit(1)

        if (not all_ok) and (not self._exit_0):
            exit(1)

        exit(0)

    def _main(self):
        """
        Base vunit main function without performing exit
        """
        if self._list_only:
            return self._main_list_only()

        if self._compile_only:
            return self._main_compile_only()

        simulator_if = self._create_simulator_if()
        test_cases = self._create_tests(simulator_if)

        self._compile(simulator_if)

        start_time = ostools.get_time()
        report = TestReport(printer=self._printer)
        try:
            self._run_test(test_cases, report)
            simulator_if.post_process(self._simulator_factory.simulator_output_path)
        except KeyboardInterrupt:
            print()
            LOGGER.debug("_main: Caught Ctrl-C shutting down")
        finally:
            del test_cases
            del simulator_if

        report.set_real_total_time(ostools.get_time() - start_time)
        self._post_process(report)

        return report.all_ok()

    def _main_list_only(self):
        """
        Main function when only listing test cases
        """
        simulator_if = None
        test_suites = self._create_tests(simulator_if)

        for test_suite in test_suites:
            for name in test_suite.test_cases:
                print(name)
        print("Listed %i tests" % test_suites.num_tests())
        return True

    def _main_compile_only(self):
        """
        Main function when only compiling
        """
        simulator_if = self._create_simulator_if()
        self._compile(simulator_if)
        return True

    def _create_output_path(self, clean):
        """
        Create or re-create the output path if necessary
        """
        if clean:
            ostools.renew_path(self._output_path)
        elif not exists(self._output_path):
            os.makedirs(self._output_path)

        ostools.renew_path(self._preprocessed_path)

    def _create_simulator_if(self):
        """
        Create a simulator interface instance
        """
        return self._simulator_factory.create()

    @property
    def vhdl_standard(self):
        return self._vhdl_standard

    @property
    def _preprocessed_path(self):
        return join(self._output_path, "preprocessed")

    @property
    def codecs_path(self):
        return join(self._output_path, "codecs")

    @property
    def use_debug_codecs(self):
        return self._use_debug_codecs

    def _create_tests(self, simulator_if):
        """
        Create the test suites by scanning the project
        """
        scanner = TestScanner(simulator_if,
                              self._configuration)
        test_list = scanner.from_project(self._project, entity_filter=self._tb_filter)

        if test_list.num_tests() == 0:
            LOGGER.warning("Test scanner found no test benches using current filter rule:\n%s",
                           self._tb_filter.__doc__)

        test_list.keep_matches(self._test_filter)
        return test_list

    def _compile(self, simulator_if):
        """
        Compile entire project
        """
        simulator_if.compile_project(self._project, self._vhdl_standard)

    def _run_test(self, test_cases, report):
        """
        Run the test suites and return the report
        """
        runner = TestRunner(report,
                            join(self._output_path, "tests"),
                            verbose=self._verbose,
                            num_threads=self._num_threads)
        runner.run(test_cases)

    def _post_process(self, report):
        """
        Print the report to stdout and optionally write it to an XML file
        """
        report.print_str()

        if self._xunit_xml is not None:
            xml = report.to_junit_xml_str()
            ostools.write_file(self._xunit_xml, xml)

    def add_builtins(self, library_name="vunit_lib", mock_lang=False, mock_log=False):
        """
        Add vunit VHDL builtin libraries
        """
        library = self.add_library(library_name)
        add_vhdl_builtins(library, self._vhdl_standard, mock_lang, mock_log)

    def add_com(self, library_name="vunit_lib", use_debug_codecs=None):
        """
        Add communication package
        """
        if not self._project.has_library(library_name):
            library = self.add_library(library_name)
        else:
            library = self.library(library_name)
        if use_debug_codecs is not None:
            self._use_debug_codecs = use_debug_codecs
        add_com(library, self._vhdl_standard,
                use_debug_codecs=self._use_debug_codecs)

    def add_array_util(self, library_name="vunit_lib"):
        """
        Add array utility package
        """
        library = self.library(library_name)
        add_array_util(library, self._vhdl_standard)

    def add_osvvm(self, library_name="osvvm"):
        """
        Add osvvm library
        """
        if not self._project.has_library(library_name):
            library = self.add_library(library_name)
        else:
            library = self.library(library_name)
        add_osvvm(library)

    def get_project_compile_order(self, target=None):
        """
        Get all project files in compile order.  An optional target
        file may be specified causing only its direct and indirect
        dependencies to be included.
        """
        if target is not None:
            target = abspath(target)
        return self._project.get_dependencies_in_compile_order(target=target)


class LibraryFacade(object):
    """
    User interface of a library
    """
    def __init__(self, library_name, parent, project, configuration):
        self._library_name = library_name
        self._parent = parent
        self._project = project
        self._configuration = configuration
        self._scope = create_scope(self._library_name)

    def set_generic(self, name, value):
        """ Set generic within library """
        self._configuration.set_generic(name.lower(), value, scope=self._scope)

    def set_parameter(self, name, value):
        """ Set generic within library  """
        self.set_generic(name, value)

    def set_sim_option(self, name, value):
        """
        Set simulation option within library
        """
        self._configuration.set_sim_option(name, value, scope=self._scope)

    def set_compile_option(self, name, value):
        """
        Set compile option within library
        """
        for source_file in self._project.get_source_files_in_order():
            if source_file.library.name == self._library_name:
                source_file.set_compile_option(name, value)

    def set_pli(self, value):
        """ Set pli within library """
        self._configuration.set_pli(value, scope=self._scope)

    def disable_ieee_warnings(self):
        """
        Disable ieee warnings within library
        """
        self._configuration.disable_ieee_warnings(scope=self._scope)

    def get_source_file(self, file_name):
        return self._parent.get_source_files(file_name, self._library_name)

    def get_source_files(self, pattern="*", allow_empty=False):
        return self._parent.get_source_files(pattern, self._library_name, allow_empty)

    def add_source_files(self, pattern, preprocessors=None, allow_empty=False):
        return self._parent.add_source_files(pattern, self._library_name, preprocessors,
                                             allow_empty=allow_empty)

    def add_source_file(self, pattern, preprocessors=None):
        return self._parent.add_source_file(pattern, self._library_name, preprocessors)

    def package(self, package_name):
        """
        Return the package with package_name or raise KeyError if does not exist
        """
        library = self._project.get_library(self._library_name)
        design_unit = library.primary_design_units.get(package_name)

        if design_unit is None:
            raise KeyError(package_name)
        if design_unit.unit_type is not 'package':
            raise KeyError(package_name)

        return PackageFacade(self._parent, self._library_name, package_name, design_unit)

    def entity(self, entity_name):
        """
        Return the entity with entity_name or raise KeyError if does not exist
        """
        library = self._project.get_library(self._library_name)
        if not library.has_entity(entity_name):
            raise KeyError(entity_name)

        return EntityFacade(self._library_name, entity_name,
                            self._configuration)

    def module(self, name):
        """
        Return the module with name or raise KeyError if does not exist
        """
        library = self._project.get_library(self._library_name)
        if name not in library.modules:
            raise KeyError(name)

        return EntityFacade(self._library_name, name,
                            self._configuration)


class EntityFacade(object):
    """
    User interface of an entity
    """
    def __init__(self, library_name, entity_name, config):
        self._library_name = library_name
        self._entity_name = entity_name
        self._config = config
        self._scope = create_scope(library_name, entity_name)

    def set_generic(self, name, value):
        """ Set generic within entity """
        self._config.set_generic(name.lower(), value, scope=self._scope)

    def set_parameter(self, name, value):
        """ Set generic within module """
        self.set_generic(name, value)

    def set_sim_option(self, name, value):
        """
        Set simulation option within entity
        """
        self._config.set_sim_option(name, value, scope=self._scope)

    def set_pli(self, value):
        """ Set pli within entity """
        self._config.set_pli(value, scope=self._scope)

    def add_config(self,  # pylint: disable=too-many-arguments
                   name="", generics=None, parameters=None, pre_config=None, post_check=None):
        """
        Add a run-configuration of all tests within entity
        """
        generics = {} if generics is None else generics
        generics = lower_generics(generics)
        parameters = {} if parameters is None else parameters
        generics.update(parameters)
        self._config.add_config(scope=self._scope,
                                name=name,
                                generics=generics,
                                pre_config=pre_config,
                                post_check=post_check)

    def disable_ieee_warnings(self):
        """
        Disable ieee warnings within entity
        """
        self._config.disable_ieee_warnings(scope=self._scope)

    def test(self, test_name):
        """
        Return the test case with test_name

        @TODO we cannot check that test exists at this point since tests are scanned in main
        @TODO assumes there is only one architecture for test benchs
        @TODO makes sense to have multiple architectures of a test bench?
        """

        return TestFacade(self._library_name, self._entity_name, test_name, self._config)


class PackageFacade(object):
    """
    User interface of a Package
    """
    def __init__(self, parent, library_name, package_name, design_unit):
        self._parent = parent
        self._library_name = library_name
        self._package_name = package_name
        self._design_unit = design_unit

    def generate_codecs(self, codec_package_name=None, used_packages=None, output_file_name=None):
        """
        Generates codecs for the datatypes in this Package
        """
        if codec_package_name is None:
            codec_package_name = self._package_name + '_codecs'

        if output_file_name is None:
            codecs_path = join(self._parent.codecs_path, self._library_name)
            file_extension = splitext(self._design_unit.source_file.name)[1]
            output_file_name = join(codecs_path, codec_package_name + file_extension)

        codec_generator.generate_codecs(self._design_unit,
                                        codec_package_name,
                                        used_packages,
                                        output_file_name,
                                        self._parent.use_debug_codecs)

        return self._parent.add_source_files(output_file_name, self._library_name)


class TestFacade(object):
    """
    User interface of a single test case

    """
    def __init__(self, library_name, entity_name, test_name, config):
        self._library_name = library_name
        self._entity_name = entity_name
        self._test_name = test_name
        self._config = config
        self._scope = create_scope(library_name, entity_name, test_name)

    def add_config(self,  # pylint: disable=too-many-arguments
                   name="", generics=None, parameters=None, pre_config=None, post_check=None):
        """
        Add a run-configuration this test case
        """
        generics = {} if generics is None else generics
        generics = lower_generics(generics)
        parameters = {} if parameters is None else parameters
        generics.update(parameters)
        self._config.add_config(scope=self._scope,
                                name=name,
                                generics=generics,
                                pre_config=pre_config,
                                post_check=post_check)

    def set_generic(self, name, value):
        """
        Set generic for test case
        """
        self._config.set_generic(name.lower(), value, scope=self._scope)

    def set_parameter(self, name, value):
        """
        Set generic within test case
        """
        self.set_generic(name, value)

    def set_sim_option(self, name, value):
        """
        Set simulation option within entity
        """
        self._config.set_sim_option(name, value, scope=self._scope)

    def disable_ieee_warnings(self):
        """
        Disable ieee warnings for test case
        """
        self._config.disable_ieee_warnings(scope=self._scope)


class FileSetFacade(list):
    """
    A set of several files
    """

    def __init__(self, source_files):
        list.__init__(self, source_files)

    def set_compile_option(self, name, value):
        """
        Set compile option
        """
        for source_file in self:
            source_file.set_compile_option(name, value)

    def depends_on(self, source_file):
        """
        Add manual dependency of these files on another
        """
        for my_source_file in self:
            my_source_file.depends_on(source_file)


class FileFacade(object):
    """
    A single file
    """
    def __init__(self, source_file, project):
        self._source_file = source_file
        self._project = project

    def set_compile_option(self, name, value):
        """
        Set compile option
        """
        self._source_file.set_compile_option(name, value)

    def depends_on(self, source_file):
        """
        Add manual dependency of this file on another
        """
        if isinstance(source_file, FileFacade):
            private_source_file = source_file._source_file  # pylint: disable=protected-access
            self._project.add_manual_dependency(self._source_file,
                                                depends_on=private_source_file)
        elif hasattr(source_file, "__iter__"):
            for element in source_file:
                self.depends_on(element)
        else:
            raise ValueError(source_file)


def select_vhdl_standard():
    """
    Select VHDL standard according to environment variable VUNIT_VHDL_STANDARD
    """
    vhdl_standard = os.environ.get('VUNIT_VHDL_STANDARD', '2008')
    assert vhdl_standard in ('93', '2002', '2008')
    return vhdl_standard


def lower_generics(generics):
    """
    Convert all generics names to lower case to match internal representation.
    @TODO Maybe warn in case of conflict. VHDL forbids this though so the user will notice anyway.
    """
    return dict((name.lower(), value) for name, value in generics.items())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Verilog preprocessing
"""

from __future__ import print_function
import argparse
from vunit.parsing.tokenizer import Tokenizer, Token, new_token_kind

TOKENIZER = Tokenizer()


def slice_value(token, start=None, end=None):
    return Token(token.kind, token.value[start:end])


def remove_value(token):
    return Token(token.kind, '')


def ignore_value(token):  # pylint: disable=unused-argument
    return None


PREPROCESSOR = TOKENIZER.add(
    "preprocessor",
    r"`[a-zA-Z][a-zA-Z0-9_]*",
    lambda token: slice_value(token, start=1))

STRING = TOKENIZER.add(
    "string",
    r'(?<!\\)"(.*?)(?<!\\)"',
    lambda token: slice_value(token, start=1, end=-1))

COMMENT = TOKENIZER.add(
    "comment",
    r'//.*$',
    lambda token: slice_value(token, start=2))


__KEYWORDS__ = {}


def new_keyword(name):
    """
    Create a new keyword
    """
    kind = new_token_kind(name)
    __KEYWORDS__[name] = kind
    return kind


def replace_keywords(token):
    if token.value in __KEYWORDS__:
        return Token(__KEYWORDS__[token.value], '')
    else:
        return token


MODULE = new_keyword("module")
ENDMODULE = new_keyword("endmodule")
PACKAGE = new_keyword("package")
IMPORT = new_keyword("import")
ENDPACKAGE = new_keyword("endpackage")
PARAMETER = new_keyword("parameter")

# Complete list of SystemVerilog keywords
__COMPLETE_KEYWORDS__ = """
accept_on
alias
always
always_comb
always_ff
always_latch
and
assert
assign
assume
automatic
before
begin
bind
bins
binsof
bit
break
buf
bufif0
bufif1
byte
case
casex
casez
cell
chandle
checker
class
clocking
cmos
config
const
constraint
context
continue
cover
covergroup
coverpoint
cross
deassign
default
defparam
design
disable
dist
do
edge
else
end
endcase
endchecker
endclass
endclocking
endconfig
endfunction
endgenerate
endgroup
endinterface
endmodule
endpackage
endprimitive
endprogram
endproperty
endspecify
endsequence
endtable
endtask
enum
event
eventually
expect
export
extends
extern
final
first_match
for
force
foreach
forever
fork
forkjoin
function
generate
genvar
global
highz0
highz1
if
iff
ifnone
ignore_bins
illegal_bins
implements
implies
import
incdir
include
initial
inout
input
inside
instance
int
integer
interconnect
interface
intersect
join
join_any
join_none
large
let
liblist
library
local
localparam
logic
longint
macromodule
matches
medium
modport
module
nand
negedge
nettype
new
nexttime
nmos
nor
noshowcancelled
not
notif0
notif1
null
or
output
package
packed
parameter
pmos
posedge
primitive
priority
program
property
protected
pull0
pull1
pulldown
pullup
pulsestyle_ondetect
pulsestyle_onevent
pure
rand
randc
randcase
randsequence
rcmos
real
realtime
ref
reg
reject_on
release
repeat
restrict
return
rnmos
rpmos
rtran
rtranif0
rtranif1
s_always
s_eventually
s_nexttime
s_until
s_until_with
scalared
sequence
shortint
shortreal
showcancelled
signed
small
soft
solve
specify
specparam
static
string
strong
strong0
strong1
struct
super
supply0
supply1
sync_accept_on
sync_reject_on
table
tagged
task
this
throughout
time
timeprecision
timeunit
tran
tranif0
tranif1
tri
tri0
tri1
triand
trior
trireg
type
typedef
union
unique
unique0
unsigned
until
until_with
untyped
use
uwire
var
vectored
virtual
void
wait
wait_order
wand
weak
weak0
weak1
while
wildcard
wire
with
within
wor
xnor
xor
""".split()

# Add keywords that we do not yet care about
for keyword in __COMPLETE_KEYWORDS__:
    if keyword not in __KEYWORDS__:
        new_keyword(keyword)

IDENTIFIER = TOKENIZER.add(
    "identifier",
    r"[a-zA-Z_][a-zA-Z0-9_]*",
    replace_keywords)

ESCAPED_NEWLINE = TOKENIZER.add(
    "escaped_newline",
    r"\\\n",
    ignore_value)

NEWLINE = TOKENIZER.add(
    "newline",
    r"\n",
    remove_value)

WHITESPACE = TOKENIZER.add(
    "whitespace",
    r"\s +",
    remove_value)

MULTI_COMMENT = TOKENIZER.add(
    "multi_line_comment",
    r"/\*(.|\n)*?\*/",
    lambda token: slice_value(token, start=2, end=-2))

SEMI_COLON = TOKENIZER.add(
    "semi_colon",
    r";",
    remove_value)

HASH = TOKENIZER.add(
    "hash",
    r"\#",
    remove_value)

EQUAL = TOKENIZER.add(
    "equal",
    r"=",
    remove_value)

LPAR = TOKENIZER.add(
    "left_parenthesis",
    r"\(",
    remove_value)

RPAR = TOKENIZER.add(
    "right_parenthesis",
    r"\)",
    remove_value)

COMMA = TOKENIZER.add(
    "comma",
    r",",
    remove_value)

OTHER = TOKENIZER.add(
    "other",
    r".+?")

TOKENIZER.finalize()


def tokenize(code):
    """
    Tokenize Verilog code to be preprocessed
    """

    return TOKENIZER.tokenize(code)


def main():
    """
    Print tokenization of a file
    """
    parser = argparse.ArgumentParser(description="Verilog preprocessor")

    parser.add_argument('file_name', nargs=1,
                        help='File name')

    args = parser.parse_args()
    with open(args.file_name[0], "r") as fptr:
        for tok in tokenize(fptr.read()):
            if tok.kind not in (WHITESPACE, NEWLINE):
                print(tok)


if __name__ == "__main__":
    main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
The command line interface of VUnit.
"""

import argparse
from os.path import join, abspath
import os
from vunit.simulator_factory import SimulatorFactory


class VUnitCLI(object):
    """
    VUnit command line interface
    """

    def __init__(self, description=None):
        self.parser = self.create_argument_parser(description)

    def parse_args(self, argv=None):
        """
        Parse arguments from 'argv' if not None instead of sys.argv
        """
        return self.parser.parse_args(args=argv)

    @staticmethod
    def create_argument_parser(description=None):
        """
        Create the argument parser
        """
        description = 'VUnit command line tool.' if description is None else description

        parser = argparse.ArgumentParser(description=description)

        parser.add_argument('test_patterns', metavar='tests', nargs='*',
                            default='*',
                            help='Tests to run')

        parser.add_argument('-l', '--list', action='store_true',
                            default=False,
                            help='Only list all test cases')

        parser.add_argument('--compile', action='store_true',
                            default=False,
                            help='Only compile project without running tests')

        parser.add_argument('--elaborate', action='store_true',
                            default=False,
                            help='Only elaborate test benches without running')

        parser.add_argument('--clean', action='store_true',
                            default=False,
                            help='Remove output path first')

        parser.add_argument('-o', '--output-path',
                            default=join(abspath(os.getcwd()), "vunit_out"),
                            help='Output path for compilation and simulation artifacts')

        parser.add_argument('-x', '--xunit-xml',
                            default=None,
                            help='Xunit test report .xml file')

        parser.add_argument('--exit-0',
                            default=False,
                            action="store_true",
                            help=('Exit with code 0 even if a test failed. '
                                  'Still exits with code 1 on fatal errors such as compilation failure'))

        parser.add_argument('-v', '--verbose', action="store_true",
                            default=False,
                            help='Print test output immediately and not only when failure')

        parser.add_argument('--no-color', action='store_true',
                            default=False,
                            help='Do not color output')

        parser.add_argument('--log-level',
                            default="warning",
                            choices=["info", "error", "warning", "debug"])

        parser.add_argument('-p', '--num-threads', type=positive_int,
                            default=1,
                            help=('Number of tests to run in parallel. '
                                  'Test output is not continuously written in verbose mode with p > 1'))

        com = parser.add_argument_group("com", description="Flags specific to the com message passing package")
        com.add_argument('--use-debug-codecs', action='store_true',
                         default=False,
                         help='Run with debug features enabled')
        SimulatorFactory.add_arguments(parser)

        return parser


def positive_int(val):
    """
    ArgumentParse positive int check
    """
    try:
        ival = int(val)
        assert ival > 0
        return ival
    except (ValueError, AssertionError):
        raise argparse.ArgumentTypeError("'%s' is not a valid positive int" % val)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2014-2016, Lars Asplund lars.anders.asplund@gmail.com

"""
Acceptance test of the VUnit public interface class
"""


import unittest
from string import Template
from os.path import join, dirname, basename, exists
import os
import re
from shutil import rmtree
from re import MULTILINE
from vunit.ui import VUnit
from vunit.project import VHDL_EXTENSIONS, VERILOG_EXTENSIONS
from vunit.test.mock_2or3 import mock
from vunit.ostools import renew_path


class TestUi(unittest.TestCase):
    """
    Testing the VUnit public interface class
    """
    def setUp(self):
        self.tmp_path = join(dirname(__file__), "test_ui_tmp")
        renew_path(self.tmp_path)
        self.cwd = os.getcwd()
        os.chdir(self.tmp_path)

        self._output_path = join(self.tmp_path, 'output')
        self._preprocessed_path = join(self._output_path, "preprocessed")
        self.mocksim = None

    def tearDown(self):
        os.chdir(self.cwd)
        if exists(self.tmp_path):
            rmtree(self.tmp_path)

    def test_global_custom_preprocessors_should_be_applied_in_the_order_they_are_added(self):
        ui = self._create_ui()
        ui.add_preprocessor(VUnitfier())
        ui.add_preprocessor(ParentalControl())

        file_name = self.create_entity_file()
        ui.add_source_files(file_name, 'lib')

        pp_source = Template("""
library vunit_lib;
context vunit_lib.vunit_context;

entity $entity is
end entity;

architecture arch of $entity is
begin
    log("Hello World");
    check_relation(1 /= 2);
    log("Here I am!"); -- VUnitfier preprocessor: Report turned of[BEEP]eeping original code.
end architecture;
""")
        with open(join(self._preprocessed_path, 'lib', basename(file_name))) as fread:
            self.assertEqual(fread.read(), pp_source.substitute(entity='ent0', file=basename(file_name)))

    def test_global_check_and_location_preprocessors_should_be_applied_after_global_custom_preprocessors(self):
        ui = self._create_ui()
        ui.enable_location_preprocessing()
        ui.enable_check_preprocessing()
        ui.add_preprocessor(TestPreprocessor())

        file_name = self.create_entity_file()
        ui.add_source_files(file_name, 'lib')

        pp_source = Template("""\
-- check_relation(a = b, line_num => 1, file_name => "$file", \
auto_msg => "Relation a = b failed! Left is " & to_string(a) & ". Right is " & to_string(b) & ".");

library vunit_lib;
context vunit_lib.vunit_context;

entity $entity is
end entity;

architecture arch of $entity is
begin
    log("Hello World", line_num => 11, file_name => "$file");
    check_relation(1 /= 2, line_num => 12, file_name => "$file", \
auto_msg => "Relation 1 /= 2 failed! Left is " & to_string(1) & ". Right is " & to_string(2) & ".");
    report "Here I am!";
end architecture;
""")
        with open(join(self._preprocessed_path, 'lib', basename(file_name))) as fread:
            self.assertEqual(fread.read(), pp_source.substitute(entity='ent0', file=basename(file_name)))

    def test_locally_specified_preprocessors_should_be_used_instead_of_any_globally_defined_preprocessors(self):
        ui = self._create_ui()
        ui.enable_location_preprocessing()
        ui.enable_check_preprocessing()
        ui.add_preprocessor(TestPreprocessor())

        file_name1 = self.create_entity_file(1)
        file_name2 = self.create_entity_file(2)

        ui.add_source_files(file_name1, 'lib', [])
        ui.add_source_files(file_name2, 'lib', [VUnitfier()])

        pp_source = Template("""
library vunit_lib;
context vunit_lib.vunit_context;

entity $entity is
end entity;

architecture arch of $entity is
begin
    log("Hello World");
    check_relation(1 /= 2);
    $report
end architecture;
""")
        self.assertFalse(exists(join(self._preprocessed_path, 'lib', basename(file_name1))))
        with open(join(self._preprocessed_path, 'lib', basename(file_name2))) as fread:
            expectd = pp_source.substitute(
                entity='ent2',
                report='log("Here I am!"); -- VUnitfier preprocessor: Report turned off, keeping original code.')
            self.assertEqual(fread.read(), expectd)

    def test_supported_source_file_suffixes(self):
        """Test adding a supported filetype, of any case, is accepted."""
        ui = self._create_ui()
        accepted_extensions = VHDL_EXTENSIONS + VERILOG_EXTENSIONS
        allowable_extensions = [ext for ext in accepted_extensions]
        allowable_extensions.extend([ext.upper() for ext in accepted_extensions])
        allowable_extensions.append(VHDL_EXTENSIONS[0][0] + VHDL_EXTENSIONS[0][1].upper() +
                                    VHDL_EXTENSIONS[0][2:])  # mixed case
        for idx, ext in enumerate(allowable_extensions):
            file_name = self.create_entity_file(idx, ext)
            ui.add_source_files(file_name, 'lib')

    def test_unsupported_source_file_suffixes(self):
        """Test adding an unsupported filetype is rejected"""
        ui = self._create_ui()
        unsupported_name = "unsupported.docx"
        self.create_file(unsupported_name)
        self.assertRaises(RuntimeError, ui.add_source_files, unsupported_name, 'lib')

    def test_can_add_non_ascii_encoded_files(self):
        ui = self._create_ui()
        lib = ui.library("lib")
        lib.add_source_files(join(dirname(__file__), 'test_ui_encoding.vhd'))
        lib.entity("encoding")  # Fill raise exception of not found

    def test_exception_on_adding_zero_files(self):
        ui = self._create_ui()
        lib = ui.library("lib")
        self.assertRaisesRegexp(ValueError, "Pattern.*missing1.vhd.*",
                                lib.add_source_files, join(dirname(__file__), 'missing1.vhd'))
        self.assertRaisesRegexp(ValueError, "File.*missing2.vhd.*",
                                lib.add_source_file, join(dirname(__file__), 'missing2.vhd'))

    def test_no_exception_on_adding_zero_files_when_allowed(self):
        ui = self._create_ui()
        lib = ui.library("lib")
        lib.add_source_files(join(dirname(__file__), 'missing.vhd'), allow_empty=True)

    def test_get_source_files(self):
        ui = self._create_ui()
        lib1 = ui.add_library("lib1")
        lib2 = ui.add_library("lib2")
        file_name = self.create_entity_file()
        lib1.add_source_file(file_name)
        lib2.add_source_file(file_name)

        self.assertEqual(len(ui.get_source_files(file_name)), 2)
        self.assertEqual(len(lib1.get_source_files(file_name)), 1)
        self.assertEqual(len(lib2.get_source_files(file_name)), 1)

        ui.get_source_file(file_name, library_name="lib1")
        ui.get_source_file(file_name, library_name="lib2")
        lib1.get_source_file(file_name)
        lib2.get_source_file(file_name)

    def test_get_source_files_errors(self):
        ui = self._create_ui()
        lib1 = ui.add_library("lib1")
        lib2 = ui.add_library("lib2")
        file_name = self.create_entity_file()
        lib1.add_source_file(file_name)
        lib2.add_source_file(file_name)
        non_existant_name = "non_existant"

        self.assertRaisesRegexp(ValueError, ".*%s.*allow_empty.*" % non_existant_name,
                                ui.get_source_files, non_existant_name)
        self.assertEqual(len(ui.get_source_files(non_existant_name, allow_empty=True)), 0)

        self.assertRaisesRegexp(ValueError, ".*named.*%s.*multiple.*library_name.*" % file_name,
                                ui.get_source_file, file_name)

        self.assertRaisesRegexp(ValueError, ".*Found no file named.*%s'" % non_existant_name,
                                ui.get_source_file, non_existant_name)

        self.assertRaisesRegexp(ValueError, ".*Found no file named.*%s.* in library 'lib1'" % non_existant_name,
                                ui.get_source_file, non_existant_name, "lib1")

    @mock.patch("vunit.project.Project.add_manual_dependency", autospec=True)
    def test_add_single_file_manual_dependencies(self, add_manual_dependency):
        # pylint: disable=protected-access
        ui = self._create_ui()
        lib = ui.library("lib")
        file_name1 = self.create_entity_file(1)
        file_name2 = self.create_entity_file(2)
        file1 = lib.add_source_file(file_name1)
        file2 = lib.add_source_file(file_name2)

        add_manual_dependency.assert_has_calls([])
        file1.depends_on(file2)
        add_manual_dependency.assert_has_calls([
            mock.call(ui._project,
                      file1._source_file,
                      depends_on=file2._source_file)])

    @mock.patch("vunit.project.Project.add_manual_dependency", autospec=True)
    def test_add_multiple_file_manual_dependencies(self, add_manual_dependency):
        # pylint: disable=protected-access
        ui = self._create_ui()
        lib = ui.library("lib")
        self.create_file("foo1.vhd")
        self.create_file("foo2.vhd")
        self.create_file("foo3.vhd")
        self.create_file("bar.vhd")
        foo_files = lib.add_source_files("foo*.vhd")
        bar_file = lib.add_source_file("bar.vhd")

        add_manual_dependency.assert_has_calls([])
        bar_file.depends_on(foo_files)
        add_manual_dependency.assert_has_calls([
            mock.call(ui._project,
                      bar_file._source_file,
                      depends_on=foo_file._source_file)
            for foo_file in foo_files])

    @mock.patch("vunit.project.Project.add_manual_dependency", autospec=True)
    def test_add_fileset_manual_dependencies(self, add_manual_dependency):
        # pylint: disable=protected-access
        ui = self._create_ui()
        lib = ui.library("lib")
        self.create_file("foo1.vhd")
        self.create_file("foo2.vhd")
        self.create_file("foo3.vhd")
        self.create_file("bar.vhd")
        foo_files = lib.add_source_files("foo*.vhd")
        bar_file = lib.add_source_file("bar.vhd")

        add_manual_dependency.assert_has_calls([])
        foo_files.depends_on(bar_file)
        add_manual_dependency.assert_has_calls([
            mock.call(ui._project,
                      foo_file._source_file,
                      depends_on=bar_file._source_file)
            for foo_file in foo_files])

    def _test_pre_config_helper(self, retval, test_not_entity=False):
        """
        Helper method to test pre_config where the pre config can return different values
        and it can be configured on both entity and test objects
        """
        ui = self._create_ui("lib.test_ui_tb.test_one")
        lib = ui.library("lib")
        lib.add_source_files(join(dirname(__file__), 'test_ui_tb.vhd'))

        if test_not_entity:
            obj = lib.entity("test_ui_tb").test("test_one")
        else:
            obj = lib.entity("test_ui_tb")

        def side_effect():
            "No simulate calls should have occured so far"
            self.assertEqual(self.mocksim.simulate_calls, [])
            return retval

        pre_config = mock.Mock()
        pre_config.return_value = retval
        pre_config.side_effect = side_effect

        obj.add_config(name="", pre_config=pre_config)
        self._run_main(ui, 0 if retval else 1)

        pre_config.assert_has_calls([call()])

    def test_entity_has_pre_config(self):
        for retval in (True, False, None):
            self._test_pre_config_helper(retval)

    def test_test_has_pre_config(self):
        self._test_pre_config_helper(True, test_not_entity=False)

    @mock.patch("vunit.ui.LOGGER", autospec=True)
    def test_warning_on_no_tests(self, logger):
        ui = self._create_ui("--compile")
        self._run_main(ui)
        self.assertEqual(logger.warning.mock_calls, [])
        logger.reset_mock()

        ui = self._create_ui("--list")
        self._run_main(ui)
        self.assertEqual(len(logger.warning.mock_calls), 1)
        self.assertTrue("found no test benches" in str(logger.warning.mock_calls))
        logger.reset_mock()

        ui = self._create_ui()
        self._run_main(ui)
        self.assertEqual(len(logger.warning.mock_calls), 1)
        self.assertTrue("found no test benches" in str(logger.warning.mock_calls))
        logger.reset_mock()

    def _create_ui(self, *args):
        """ Create an instance of the VUnit public interface class """
        ui = VUnit.from_argv(argv=["--output-path=%s" % self._output_path,
                                   "--clean"] + list(args),
                             compile_builtins=False)
        ui.add_library('lib')

        factory = MockSimulatorFactory(self._output_path)
        self.mocksim = factory.mocksim
        ui._simulator_factory = factory  # pylint: disable=protected-access
        return ui

    def _run_main(self, ui, code=0):
        """
        Run ui.main and expect exit code
        """
        try:
            ui.main()
        except SystemExit as exc:
            self.assertEqual(exc.code, code)

    def create_entity_file(self, idx=0, file_suffix='.vhd'):
        """
        Create and a temporary file containing the same source code
        but with different entity names depending on the index
        """
        source = Template("""
library vunit_lib;
context vunit_lib.vunit_context;

entity $entity is
end entity;

architecture arch of $entity is
begin
    log("Hello World");
    check_relation(1 /= 2);
    report "Here I am!";
end architecture;
""")

        entity_name = "ent%i" % idx
        file_name = entity_name + file_suffix
        self.create_file(file_name,
                         source.substitute(entity=entity_name))
        return file_name

    @staticmethod
    def create_file(file_name, contents=""):
        """
        Creata file in the temporary path with given contents
        """
        with open(file_name, "w") as fptr:
            fptr.write(contents)


def call(*args, **kwargs):
    """
    To create call objects for checking calls on mock objects
    """
    return (args, kwargs)


class TestPreprocessor(object):
    """
    A preprocessor that appends a check_relation call before the orginal code
    """
    def __init__(self):
        pass

    @staticmethod
    def run(code, file_name):  # pylint: disable=unused-argument
        return '-- check_relation(a = b);\n' + code


class VUnitfier(object):
    """
    A preprocessor that replaces report statments with log calls
    """
    def __init__(self):
        self._report_pattern = re.compile(r'^(?P<indent>\s*)report\s*(?P<note>"[^"]*")\s*;', MULTILINE)

    def run(self, code, file_name):   # pylint: disable=unused-argument
        return self._report_pattern.sub(
            r'\g<indent>log(\g<note>); -- VUnitfier preprocessor: Report turned off, keeping original code.', code)


class ParentalControl(object):
    """
    A preprocessor that replaces f..k with [BEEP]
    """
    def __init__(self):
        self._fword_pattern = re.compile(r'f..k')

    def run(self, code, file_name):   # pylint: disable=unused-argument
        return self._fword_pattern.sub(r'[BEEP]', code)


class MockSimulatorFactory(object):
    """
    Mock a simulator factory
    """
    simulator_name = "mocksim"

    def __init__(self, output_path):
        self._output_path = output_path
        self.mocksim = MockSimulator()

    @property
    def simulator_output_path(self):
        return join(self._output_path, self.simulator_name)

    @staticmethod
    def package_users_depend_on_bodies():
        return True

    def create(self):
        return self.mocksim


class MockSimulator(object):
    """
    Mock a simulator
    """

    def __init__(self):
        self.simulate_calls = []

    @staticmethod
    def compile_project(project, vhdl_standard):  # pylint: disable=unused-argument
        return True

    def post_process(self, output_path):  # pylint: disable=unused-argument
        pass

    def simulate(self,   # pylint: disable=too-many-arguments
                 output_path, library_name, entity_name, architecture_name, config):
        """
        Accumulate simulation calls
        """
        self.simulate_calls.append((output_path,
                                    library_name,
                                    entity_name,
                                    architecture_name,
                                    config))
        return True



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
#
# Copyright (c) 2015, Lars Asplund lars.anders.asplund@gmail.com

"""
Test of the Verilog preprocessor
"""

from os.path import join, dirname, exists
import os
from vunit.ostools import renew_path

from unittest import TestCase
from vunit.parsing.verilog.preprocess import preprocess, Macro
from vunit.parsing.verilog.tokenizer import tokenize


class TestVerilogPreprocessor(TestCase):
    """
    Test of the Verilog preprocessor
    """

    def setUp(self):
        self.output_path = join(dirname(__file__), "test_verilog_preprocessor_out")
        renew_path(self.output_path)

    def test_non_preprocess_tokens_are_kept(self):
        defines = {}
        tokens = tokenize('"hello"ident/*comment*///comment')
        pp_tokens = preprocess(tokenize('"hello"ident/*comment*///comment'), defines)
        self.assertEqual(pp_tokens, tokens)
        self.assertEqual(defines, {})

    def test_preprocess_define_without_value(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo"), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines, {"foo": Macro("foo")})

    def test_preprocess_define_with_value(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo bar \"abc\""), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines, {"foo": Macro("foo", tokenize("bar \"abc\""))})

    def test_preprocess_define_with_lpar_value(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo (bar)"), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines, {"foo": Macro("foo", tokenize("(bar)"))})

    def test_preprocess_define_with_one_arg(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo(arg)arg 123"), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines,
                         {"foo": Macro("foo", tokenize("arg 123"), args=("arg",))})

    def test_preprocess_define_with_multiple_args(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo( arg1, arg2)arg1 arg2"), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines,
                         {"foo": Macro("foo", tokenize("arg1 arg2"), args=("arg1", "arg2"))})

    def test_preprocess_define_with_default_values(self):
        defines = {}
        tokens = preprocess(tokenize("`define foo(arg1, arg2=default)arg1 arg2"), defines)
        self.assertEqual(tokens, [])
        self.assertEqual(defines,
                         {"foo": Macro("foo",
                                       tokenize("arg1 arg2"),
                                       args=("arg1", "arg2"),
                                       defaults={"arg2": tokenize("default")})})

    def test_preprocess_substitute_define_without_args(self):
        tokens = preprocess(tokenize("""\
`define foo bar \"abc\"
`foo"""))
        self.assertEqual(tokens, tokenize("bar \"abc\""))

    def test_preprocess_substitute_define_with_one_arg(self):
        tokens = preprocess(tokenize("""\
`define foo(arg)arg 123
`foo(hello hey)"""))
        self.assertEqual(tokens, tokenize("hello hey 123"))

    def test_preprocess_substitute_define_with_multile_args(self):
        tokens = preprocess(tokenize("""\
`define foo(arg1, arg2)arg1,arg2
`foo(1 2, hello)"""))
        self.assertEqual(tokens, tokenize("1 2, hello"))

    def test_preprocess_substitute_define_with_default_values(self):
        defines = {}
        tokens = preprocess(tokenize("""\
`define foo(arg1, arg2=default)arg1 arg2
`foo(1)"""), defines)
        self.assertEqual(tokens, tokenize("1 default"))

    def test_preprocess_include_directive(self):
        self.write_file("include.svh", "hello hey")
        included_files = []
        tokens = preprocess(tokenize('`include "include.svh"'),
                            include_paths=[self.output_path],
                            included_files=included_files)
        self.assertEqual(tokens, tokenize("hello hey"))
        self.assertEqual(included_files, [join(self.output_path, "include.svh")])

    def write_file(self, file_name, contents):
        """
        Write file with contents into output path
        """
        full_name = join(self.output_path, file_name)
        full_path = dirname(full_name)
        if not exists(full_path):
            os.makedirs(dirname(full_path))
        with open(full_name, "w") as fptr:
            fptr.write(contents)
